""" Class for evaluating (S)SNMF models. """


import numpy as np
from numpy import linalg as la
import scipy.optimize.nnls as nnls

import ssnmf.ssnmf


class Evaluation:
    """
    Class for evaluating (S)SNMF models on train and test data. It contains functions
    to compute the reconstruction errors and classifications accuracies for each model
    on the train and test data.

    Model Numbers
    ----------
    (3) ||X - AS||_F^2 + lam * ||Y - BS||_F^2 or
    (4) ||X - AS||_F^2 + lam * D(Y||BS) or
    (5) D(X||AS) + lam * ||Y - BS||_F^2 or
    (6) D(X||AS) + lam * D(Y||BS)

    Parameters
    ----------
    train_features : array
                     Train data matrix, shape (#features, #train samples)
    test_features  : array
                     Test data matrix, shape (#features, #test samples)
    train_labels   : array
                     Train label matrix, shape (#classes, #train samples)
    test_labels    : array
                     Test label matrix, shape (#classes, #test samples)
    k              : int_
                     Number of topics.
    modelNum       : int_
                     Number indicating which of the (S)SNMF models to train (see model numbers above).
    A              : array, optional
                     Initialization for the left factor matrix of the data matrix, shape (#features, k)
                     (the default is a matrix with uniform random entries).
    S              : array, optional
                     Initialization for the right factor matrix of the train data/label matrix,
                     shape (k, #train samples), (the default is a matrix with uniform random entries).
    B              : array, optional
                     Initialization for left factor matrix of the train label matrix, shape (#classes, k),
                     (the default is a matrix with uniform random entries).
    lam            : float_, optional
                     Weight parameter for classification term in objective (the default is 1).
    numiters       : int_,optional
                     Number of iterations of updates to run (default is 10).
    W_train        : array, optional
                     Missing data indicator matrix for the training process, shape (#features, #train samples + #test samples)
                     i.e. the matrix is used to indicate missing entries in the train data, and indicate which test data to mask
                     in the training process (the default is matrix of all ones for train data and all zeros for test data).
    W_test         : array, optional
                     Missing data indicator matrix used in the testing process, shape (#features, #test samples),
                     i.e. the matrix is used to indicate missing entries in the test data,
                     (the default is matrix of all ones for test data).
    L              : array, optional
                     Missing label indicator matrix for the training process, shape (#classes, #train samples + #test samples)
                     (the default is matrix of all ones for train data and all zeros for test data).
    tol            : float_, optional
                     tolerance for terminating the model (default is 1e-4).
    iter_s         : int_, optional
                     Number of iterations of updates to run to approximate the representation of the test
                     data when the I-divervene a discrepancy measure for data reconstruction (default is 10).


    Methods
    ----------
    eval()
        Fits (S)SNMF model to the train data, and evaluates the performance on the train and test data.
    test_model()
        Compute the representation of the test data given a trained model.
    computeErrors(W)
        Compute the errors of the model on test data.
    iterativeLocalSearch()
        Perform an iterative local search in hyperparameter space for a set of (locally) optimal (k,lambda,iterations).


    Usage Example
    ----------
    import evaluation
    from evaluation import Evaluation

    # Run hyperparameter search (Optional)
    param_module = Evaluation(train_features = X_train,
                                 train_labels = y_train,
                                 test_features = X_val,
                                 test_labels = y_val,
                                 k = 14, modelNum = 4)
    [opt_lamb, opt_ka, opt_itas] = param_module.iterativeLocalSearch(init_lamb=100, init_k=14, init_itas=8)

    # Run model with "optimal" parameters.
    eval_module = Evaluation(train_features = np.concatenate((X_train, X_val), axis=1),
                                 train_labels = np.concatenate((y_train, y_val), axis=1),
                                 test_features = X_test,
                                 test_labels = y_test,
                                 modelNum = 4, k = opt_ka,
                                 lam=opt_lamb, numiters = opt_itas)

    train_evals, test_evals = eval_module.eval()

    print(f"The classification accuracy on the train data is {train_evals[-1][-1]*100:.2f}%")
    print(f"The classification accuracy on the test data is {test_evals[-1]*100:.2f}%")

    """


    def __init__(self, train_features, train_labels, test_features, test_labels, modelNum, k, **kwargs):
        self.train_features = train_features
        self.train_labels = train_labels
        self.test_features = test_features
        self.test_labels = test_labels
        self.modelNum = modelNum
        self.k = k

        if train_features.shape[1] != train_labels.shape[1]:
            raise Exception('The column dimensions of train_features and train_labels are not equal.')
        if test_features.shape[1] != test_labels.shape[1]:
            raise Exception('The column dimensions of test_features and test_labels are not equal.')

        self.rows = np.shape(train_features)[0] #number of features
        self.cols = np.shape(train_labels)[1] + np.shape(test_labels)[1] #number of samples in the train and test set
        self.classes = np.shape(self.train_labels)[0] #number of classes in the data
        self.lam = kwargs.get('lam',1)
        self.numiters = kwargs.get('numiters',10)
        self.tol = kwargs.get('tol', 1e-4)
        self.iter_s = kwargs.get('iter_s', 20)
        self.A = kwargs.get('A',np.random.rand(self.rows,k)) #initialize factor A
        self.S = kwargs.get('S',np.random.rand(k,self.cols)) #initialize factor S
        self.B = kwargs.get('B', np.random.rand(self.classes, k)) #initialize factor B

        if self.rows != np.shape(self.A)[0]:
            raise Exception('The row dimension of A is not equal to the number of features in the data.')
        if self.k != np.shape(self.A)[1]:
            raise Exception('The column dimension of A is not equal to k.')
        if self.k != np.shape(self.S)[0]:
            raise Exception('The row dimension of S is not equal to k.')
        if self.cols != np.shape(self.S)[1]:
            raise Exception('The column dimension of S is not equal to the number of samples in the train and test data.')
        if np.shape(self.B)[0] != self.classes:
            raise Exception('The row dimension B is not equal not equal to the number of classes.')
        if np.shape(self.B)[1] != self.k:
            raise Exception('The column dimension of B is not k.')

        self.W_train = kwargs.get('W_train', np.concatenate((np.ones(self.train_features.shape),np.zeros(self.test_features.shape)),axis=1))
        if np.shape(self.W_train)[0] != self.rows:
            raise Exception('The row dimension W_train is not equal to the number of features in the data.')
        if np.shape(self.W_train)[1] != self.cols:
            raise Exception('The column dimension of W_train is not equal to the number of samples in the train and test data.')

        self.W_test = kwargs.get('W_test', np.ones(self.test_features.shape))
        if np.shape(self.W_test)[0] != self.rows:
            raise Exception('The row dimension W_test is not equal to the number of features in the data.')
        if np.shape(self.W_test)[1] != self.test_features.shape[1]:
            raise Exception('The column dimension of W_test is not equal to the number of samples in the test data.')

        self.L = kwargs.get('L', np.concatenate((np.ones(self.train_labels.shape),np.zeros(self.test_labels.shape)),axis=1))
        if np.shape(self.L)[0] != self.classes:
            raise Exception('The row dimension of L is not equal to the number of classes in the data.')
        if np.shape(self.L)[1] != self.cols:
            raise Exception('The column dimension of L is not equal to the number of samples in the train and test data.')

        self.model = ssnmf.SSNMF(X = np.concatenate((self.train_features, self.test_features), axis = 1), k = self.k, \
                                        modelNum = self.modelNum, Y = np.concatenate((self.train_labels, self.test_labels), axis = 1), \
                                        W = self.W_train, L = self.L, A = self.A, B = self.B, S = self.S, lam = self.lam, tol = self.tol)

    def eval(self):
        '''
        This function fits the (S)SNMF model to the train data, and evaluates the performance of the model on the train and test data.

        Returns:
            train_evals (list): [ndarray(total reconstruction error (model objective function) on train data for each iteration),
                                 ndarray(data reconstruction error on train data for each iteration),
                                 ndarray(label reconstruction error on train data for each iteration),
                                 ndarray(classification accuracy on train data for each iteration)]
            test_evals (list): [float(total reconstruction error (model objective function) on test data),
                                float(data reconstruction error on test data),
                                float(label reconstruction error on test data),
                                float(classification accuracy on test data)]
        '''
        # Fit (S)SNMF model to train data
        train_evals = self.model.mult(numiters = self.numiters, saveerrs = True) #save train data errors
        # Apply (S)SNMF model to test data
        self.S_test = self.test_model() #representation matrix of test data
        # Compute reconstruction errors on test data
        test_evals = self.computeErrors()

        return train_evals, test_evals


    def test_model(self):
        '''
        Given a trained (S)SNMF model i.e. learned data dictionary, A, the function applies the (S)SNMF model
        to the test data to produce the representation of the test data.

        Returns:
            S_test (ndarray): representation matrix of the test data, shape(#topics, #test features)
        '''
        if self.modelNum == 3 or self.modelNum == 4: # Frobenius discrepancy measure on features data (use nonnegative least squares)
            S_test = np.zeros([self.k, np.shape(self.test_features)[1]])
            for i in range(np.shape(self.test_features)[1]):
                S_test[:,i] = nnls(np.multiply(np.ones(self.model.A.shape) * self.W_test[:,i].reshape(-1,1),self.model.A), \
                                    np.multiply(self.W_test[:,i], self.test_features[:,i]))[0]

        if self.modelNum == 5 or self.modelNum == 6: # I-divergence discrepancy measure on features data (use mult. upd. I-div)
            S_test = np.random.rand(self.k, self.test_features.shape[1])
            for i in range(self.iter_s):
                S_test = np.transpose(self.model.dictupdateIdiv(np.transpose(self.test_features), np.transpose(S_test), \
                                       np.transpose(self.model.A), np.transpose(self.W_test), eps= 1e-10))
        return S_test


    def computeErrors(self):
        """
        Compute errors and classification accuracy of the model on the test data.

        Returns:
            errs (float): total reconstruction error on test data (model objective function)
            reconerrs (float): data reconstruction error on test data
            classerrs (float): label reconstruction error on test data
            classaccs (float): classification accuracy on test data

        """
        L_test = np.ones(self.test_labels.shape, dtype = float) # missing label indicator matrix of same size as test_features (the default is matrix of all ones)

        if self.modelNum == 3:
            reconerrs = self.model.fronorm(self.test_features, self.model.A, self.S_test, M = self.W_test)
            classerrs = self.model.fronorm(self.test_labels, self.model.B, self.S_test, M = L_test)
            errs = reconerrs ** 2 + self.lam * classerrs ** 2
            classaccs = self.model.accuracy(Y=self.test_labels, B=self.model.B, S=self.S_test, L= L_test)
        if self.modelNum == 4:
            reconerrs = self.model.fronorm(self.test_features, self.model.A, self.S_test, M = self.W_test)
            classerrs = self.model.Idiv(self.test_labels, self.model.B, self.S_test, M = L_test)
            errs = reconerrs ** 2 + self.lam * classerrs
            classaccs = self.model.accuracy(Y=self.test_labels, B=self.model.B, S=self.S_test, L= L_test)
        if self.modelNum == 5:
            reconerrs = self.model.Idiv(self.test_features, self.model.A, self.S_test, M = self.W_test)
            classerrs = self.model.fronorm(self.test_labels, self.model.B, self.S_test, M = L_test)
            errs = reconerrs + self.lam * (classerrs ** 2)
            classaccs = self.model.accuracy(Y=self.test_labels, B=self.model.B, S=self.S_test, L = L_test)
        if self.modelNum == 6:
            reconerrs = self.model.Idiv(self.test_features, self.model.A, self.S_test, M = self.W_test)
            classerrs = self.model.Idiv(self.test_labels, self.model.B, self.S_test, M = L_test)
            errs = reconerrs + self.lam * classerrs
            classaccs = self.model.accuracy(Y=self.test_labels, B=self.model.B, S=self.S_test, L= L_test)

        return [errs, reconerrs, classerrs, classaccs]

    def iterativeLocalSearch(self, init_lamb, init_k, init_itas):
        """
        This function allows the user to perform an iterative local search in hyperparameter space for
        a set of [locally] optimal (k,lambda,iterations). This is done by starting with a choice
        of rank (init_k), iterations (init_itas), and lambda (init_lamb). Then by performing a parameter
        sweep in a "local search cube", the algorithm can determine if a new choice of (k,lambda,iterations)
        is advantageous i.e. produces better classification accuracy on validation set.

        Several hard-coded choices were made to ease the user-required knowledge of the search algorithm itself.
        These are for example:
            - The size of the local search box is (k-1 , k+2) x (iter-1, iter+1) x (0.1*lam , 10*lam).
            - The number of iterations tested uniformly spaced in the interval (iter-1, iter+1), which currently is 2 values.
            - The number of lambdas tested uniformly spaced in the interval (0.1*lam, 10*lam), which currently is 3 values.

        Args:
            init_lamb (float): initialization of the regualrizer lambda
            init_k (int): initialization of the rank
            init_itas (int): initialization of the number of iterations

        Returns:
            lamb (float): optimal regualrizer lambda found in the local search
            ka (int): optimal number of topics found in the local search
            itas (int): optimal number of iterations found in the local search
        """

        self.lam = init_lamb
        self.k = init_k
        self.numiters = init_itas
        self.converged = False
        self.iteration = 1

        if self.k <2:
            raise Exception('The initialization of the number of topics should be at least 2.')

        if self.numiters <2:
            raise Exception('The initialization of the number of iterations should be at least 2.')

        # Hyper-hyperparameters
        self.max_search_iter = 10     # Maximum number of local parameter sweeps performed
        self.ww              = 0.001   # Threshold for convergence - continue as long as a 0.1% improvement in accuracy.

        # Get the initial error
        eval_module = Evaluation(train_features = self.train_features, train_labels = self.train_labels,\
                                test_features = self.test_features, test_labels = self.test_labels,\
                                modelNum = self.modelNum, k = self.k, lam=self.lam, numiters = self.numiters,\
                                W_train = self.W_train, W_test= self.W_test,\
                                L = self.L, tol = self.tol, iter_s = self.iter_s)

        train_evals, test_evals = eval_module.eval()
        self.cls_last = test_evals[-1]

        #print(f"Initial hyperparameters: k = {self.k}, iter = {self.numiters}, and lam = {self.lam}.")
        print("Initial hyperparameters: k = {}, iter = {}, and lam = {}.".format(self.k,self.numiters,self.lam))
        #print(f"Initial val total reconstruction error = {test_evals[0]}")
        print("Initial val total reconstruction error = {}".format(test_evals[0]))
        #print(f"Initial val data reconstruction error = {test_evals[1]}")
        print("Initial val data reconstruction error = {}".format(test_evals[1]))
        #print(f"Initial val classification error = {test_evals[2]}")
        print("Initial val classification error = {}".format(test_evals[2]))
        #print(f"Initial val classification accuracy: {test_evals[3]}")
        print("Initial val classification accuracy: {}".format(test_evals[3]))

        while(not self.converged):
            # Perform a local search on k, itas, and lam
            best_ka = init_k
            best_itas = init_itas
            best_lamb = init_lamb
            best_local_acc = 0

            for k in range(self.k-1,self.k+2):
                for it in range(self.numiters-1,self.numiters+2,1):
                    startl = float(0.1*self.lam)
                    endl = float(10*self.lam)
                    la_vals = np.concatenate((np.linspace(startl, endl, num=2), np.array([self.lam])), axis=0)
                    for la_idx in range(len(la_vals)):
                        la = la_vals[la_idx]
                        #print(f"Currently testing k = {k}, iteration = {it}, and lambda = {la}")
                        print("Currently testing k = {}, iteration = {}, and lambda = {}".format(k,it,la))
                        eval_module = Evaluation(train_features = self.train_features,train_labels = self.train_labels,\
                                                test_features = self.test_features, test_labels = self.test_labels,\
                                                modelNum = self.modelNum, k = k, lam=la, numiters = it, \
                                                W_train = self.W_train, W_test= self.W_test,\
                                                L = self.L, tol = self.tol, iter_s = self.iter_s)
                        train_evals, test_evals = eval_module.eval()
                        if(test_evals[-1] > best_local_acc):
                            best_ka = k
                            best_itas = it
                            best_lamb = la
                            best_local_acc = test_evals[-1]

            # Recover the performance of the best (k,lamb,iter) in the local search box
            eval_module = Evaluation(train_features = self.train_features, train_labels = self.train_labels,\
                                    test_features = self.test_features, test_labels = self.test_labels,\
                                    modelNum = self.modelNum, k = best_ka, lam=best_lamb, numiters = best_itas,\
                                    W_train = self.W_train, W_test= self.W_test,\
                                    L = self.L, tol = self.tol, iter_s = self.iter_s)
            train_evals, test_evals = eval_module.eval()
            cls_current = test_evals[-1]

            # Compare the best performing (k,lamb,iter) with the previous best choice
            if(cls_current >= (1 + self.ww)*self.cls_last):
                self.converged  = False
                self.cls_last = cls_current
                self.k = best_ka
                self.numiters = best_itas
                self.lam = best_lamb
                self.iteration += 1

                if(self.iteration % 1 == 0):
                    print("Iteration:",self.iteration)
                    print("The new selection of hyperparameters: k =", self.k, "; iter =", self.numiters, "; lamb =", self.lam)
                    #print(f"Current val total reconstruction error = {test_evals[0]}")
                    print("Current val total reconstruction error = {}".format(test_evals[0]))
                    #print(f"Current val data reconstruction error = {test_evals[1]}")
                    print("Current val data reconstruction error = {}".format(test_evals[1]))
                    #print(f"Current val classification (div) error = {test_evals[2]}")
                    print("Current val classification (div) error = {}".format(test_evals[2]))
                    #print(f"Current val classification accuracy: {test_evals[3]}")
                    print("Current val classification accuracy: {}".format(test_evals[3]))

                if self.k  < 2:
                    self.converged = True
                    print('The search method stopped because the number of topics found is less than two.')
                    return
                if self.numiters  < 2:
                    self.converged = True
                    print('The search method stopped because the number of iterations found is less than two.')
                    return

            elif(self.iteration > self.max_search_iter):
                self.converged = True
                print("The maximum number of search iterations is reached. The set of hyperparameters found are:")
                print("k =", self.k, "; iter =", self.numiters, "; lamb =", self.lam)

            else:
                self.converged = True
                print("Converged!")
                print("The set of hyperparameters found at convergence were:")
                print("k =", self.k, "; iter =", self.numiters, "; lam =", self.lam)
                #print(f"Final val classification accuracy: {self.cls_last}")
                print("Final val classification accuracy: {}".format(self.cls_last))


        return self.lam, self.k, self.numiters



"""
Dummy conftest.py for python_boilerplate.

If you don't know what this is for, just leave it empty.
Read more about conftest.py under:
https://pytest.org/latest/plugins.html

https://stackoverflow.com/questions/34466027/in-pytest-what-is-the-use-of-conftest-py-files
"""

# import pytest


# Import necessary packages
import numpy as np
import torch
#import torchvision
import matplotlib.pyplot as plt
from time import time
import os
#from google.colab import drive
import scipy.optimize.nnls as nnls
from numpy import linalg as la

import ssnmf

def topic_plot(A, vertpixels, horizpixels, colnum):
    topic = np.transpose(np.reshape(A[:,colnum],[horizpixels,vertpixels]))
    plt.imshow(topic, cmap='binary')
    plt.show()

def visualize_reconstr(dictionary, representation, vertpixels, horizpixels, indices):
    # Indices is a list
    recon = np.matmul(dictionary,representation)
    for index in indices:
        image = np.transpose(np.reshape(recon[:,index],[horizpixels,vertpixels]))
        plt.imshow(image, cmap='binary')
        plt.savefig('./'+'reconstruction'+str(index)+'.png')



def plot_util(kchoices,train_errs,test_errs,train_reconerrs,test_reconerrs,train_classerrs,test_classerrs,train_accs,test_accs,namme,iOption):
    # iOption - indicate which x-axis label to use - options are 'k','l','n'
    # namme   - Name of figure to be saved. 

    fig, axs = plt.subplots(2, 2, figsize=(17,10))
    axs[0, 0].plot(kchoices,train_errs,color='blue',linewidth=4,label='Train Errors')
    axs[0, 0].plot(kchoices,test_errs,color='red',linestyle='dashed',linewidth=4,label='Test Errors')
    axs[0, 0].legend()
    axs[0, 0].set_title('Train Errors')

    axs[0, 1].plot(kchoices,train_reconerrs,color='blue',linewidth=4,label='Train Reconstruction Errors')
    axs[0, 1].plot(kchoices,test_reconerrs,color='red',linestyle='dashed',linewidth=4,label='Test Reconstruction Errors')
    axs[0, 1].legend()
    axs[0, 1].set_title('Train Reconstruction Errors')

    axs[1, 0].plot(kchoices,train_classerrs,color='blue',linewidth=4,label='Train Classification Errors')
    axs[1, 0].plot(kchoices,test_classerrs,color='red',linestyle='dashed',linewidth=4,label='Test Classification Errors')
    axs[1, 0].legend()
    axs[1, 0].set_title('Train Classification Errors')
    
    axs[1, 1].plot(kchoices,train_accs,color='blue',linewidth=4,label='Train Classification Accuracies')
    axs[1, 1].plot(kchoices,test_accs,color='red',linestyle='dashed',linewidth=4,label='Test Classification Accuracies')
    axs[1, 1].legend()
    axs[1, 1].set_title('Train Classification Accuracies')

    if(iOption == 'k'):
        axs[0, 0].set_xlabel('k')
        axs[0, 1].set_xlabel('k')
        axs[1, 0].set_xlabel('k')
        axs[1, 1].set_xlabel('k')

    if(iOption == 'l'):
        axs[0, 0].set_xlabel('lambda')
        axs[0, 1].set_xlabel('lambda')
        axs[1, 0].set_xlabel('lambda')
        axs[1, 1].set_xlabel('lambda')

    if(iOption == 'n'):
        axs[0, 0].set_xlabel('Iterations')
        axs[0, 1].set_xlabel('Iterations')
        axs[1, 0].set_xlabel('Iterations')
        axs[1, 1].set_xlabel('Iterations')

    lines, labels = fig.axes[-1].get_legend_handles_labels()
    fig.legend(lines, labels, loc = 'upper center')
    plt.savefig('./'+namme+'.png')


def k_plots(train_features, train_labels, test_features, test_labels, kchoices, lam, numiters,avgnum):
    numks = np.shape(kchoices)[0]
    train_errs = [0]*numks
    train_reconerrs = [0]*numks
    train_classerrs = [0]*numks
    train_accs = [0]*numks
    test_errs = [0]*numks
    test_reconerrs = [0]*numks
    test_classerrs = [0]*numks
    test_accs = [0]*numks

    for i in range(numks):
        for j in range(avgnum):
            module = TrainTestSetEvaluation(train_features, train_labels, test_features, test_labels, kchoices[i], lam, numiters)
            train_model_error, train_acc, numiters, [test_err, test_reconerr, test_classerr, test_acc], S_test = module.tt_eval_ssnmfmult()
            train_model_errors = train_model_error[0]
            train_errs[i] = train_errs[i]+train_model_errors[numiters-1]
            train_model_reconerrs = train_model_error[1]
            train_reconerrs[i] = train_reconerrs[i]+train_model_reconerrs[numiters-1]
            train_model_classerrs = train_model_error[2]
            train_classerrs[i] = train_classerrs[i]+train_model_classerrs[numiters-1]
            train_accs[i] = train_accs[i]+train_acc
            test_errs[i] = test_errs[i]+test_err
            test_reconerrs[i] = test_reconerrs[i]+test_reconerr
            test_classerrs[i] = test_classerrs[i]+test_classerr
            test_accs[i] = test_accs[i]+test_acc

    train_errs = [element/avgnum for element in train_errs]
    train_reconerrs = [element/avgnum for element in train_reconerrs]
    train_classerrs = [element/avgnum for element in train_classerrs]
    train_accs = [element/avgnum for element in train_accs]
    test_errs = [element/avgnum for element in test_errs]
    test_reconerrs = [element/avgnum for element in test_reconerrs]
    test_classerrs = [element/avgnum for element in test_classerrs]
    test_accs = [element/avgnum for element in test_accs]

    plot_util(kchoices,train_errs,test_errs,train_reconerrs,test_reconerrs,train_classerrs,test_classerrs,train_accs,test_accs,'k_SSNMF','k')

def kl_k_plots(train_features, train_labels, test_features, test_labels, kchoices, lam, numiters,avgnum):
    numks = np.shape(kchoices)[0]
    train_errs = [0]*numks
    train_reconerrs = [0]*numks
    train_classerrs = [0]*numks
    train_accs = [0]*numks
    test_errs = [0]*numks
    test_reconerrs = [0]*numks
    test_classerrs = [0]*numks
    test_accs = [0]*numks

    for i in range(numks):
        for j in range(avgnum):
            module = TrainTestSetEvaluation(train_features, train_labels, test_features, test_labels, kchoices[i], lam, numiters)
            train_model_error, train_acc, numiters, [test_err, test_reconerr, test_classerr, test_acc], S_test = module.tt_eval_kl_ssnmfmult()
            train_model_errors = train_model_error[0]
            train_errs[i] = train_errs[i]+train_model_errors[numiters-1]
            train_model_reconerrs = train_model_error[1]
            train_reconerrs[i] = train_reconerrs[i]+train_model_reconerrs[numiters-1]
            train_model_classerrs = train_model_error[2]
            train_classerrs[i] = train_classerrs[i]+train_model_classerrs[numiters-1]
            train_accs[i] = train_accs[i]+train_acc
            test_errs[i] = test_errs[i]+test_err
            test_reconerrs[i] = test_reconerrs[i]+test_reconerr
            test_classerrs[i] = test_classerrs[i]+test_classerr
            test_accs[i] = test_accs[i]+test_acc

    train_errs = [element/avgnum for element in train_errs]
    train_reconerrs = [element/avgnum for element in train_reconerrs]
    train_classerrs = [element/avgnum for element in train_classerrs]
    train_accs = [element/avgnum for element in train_accs]
    test_errs = [element/avgnum for element in test_errs]
    test_reconerrs = [element/avgnum for element in test_reconerrs]
    test_classerrs = [element/avgnum for element in test_classerrs]
    test_accs = [element/avgnum for element in test_accs]

    plot_util(kchoices,train_errs,test_errs,train_reconerrs,test_reconerrs,train_classerrs,test_classerrs,train_accs,test_accs,'k_KLSSNMF','k')
    
def lam_plots(train_features, train_labels, test_features, test_labels, k, lamchoices, numiters,avgnum):
    numlams = np.shape(lamchoices)[0]
    train_errs = [0]*numlams
    train_reconerrs = [0]*numlams
    train_classerrs = [0]*numlams
    train_accs = [0]*numlams
    test_errs = [0]*numlams
    test_reconerrs = [0]*numlams
    test_classerrs = [0]*numlams
    test_accs = [0]*numlams

    for i in range(numlams):
        for j in range(avgnum):
            module = TrainTestSetEvaluation(train_features, train_labels, test_features, test_labels, k, lamchoices[i], numiters)
            train_model_error, train_acc, numiters, [test_err, test_reconerr, test_classerr, test_acc], S_test = module.tt_eval_ssnmfmult()
            train_model_errors = train_model_error[0]
            train_errs[i] = train_errs[i]+train_model_errors[numiters-1]
            train_model_reconerrs = train_model_error[1]
            train_reconerrs[i] = train_reconerrs[i]+train_model_reconerrs[numiters-1]
            train_model_classerrs = train_model_error[2]
            train_classerrs[i] = train_classerrs[i]+train_model_classerrs[numiters-1]
            train_accs[i] = train_accs[i]+train_acc
            test_errs[i] = test_errs[i]+test_err
            test_reconerrs[i] = test_reconerrs[i]+test_reconerr
            test_classerrs[i] = test_classerrs[i]+test_classerr
            test_accs[i] = test_accs[i]+test_acc

    train_errs = [element/avgnum for element in train_errs]
    train_reconerrs = [element/avgnum for element in train_reconerrs]
    train_classerrs = [element/avgnum for element in train_classerrs]
    train_accs = [element/avgnum for element in train_accs]
    test_errs = [element/avgnum for element in test_errs]
    test_reconerrs = [element/avgnum for element in test_reconerrs]
    test_classerrs = [element/avgnum for element in test_classerrs]
    test_accs = [element/avgnum for element in test_accs]

    plot_util(lamchoices,train_errs,test_errs,train_reconerrs,test_reconerrs,train_classerrs,test_classerrs,train_accs,test_accs,'l_SSNMF','l')

def kl_lam_plots(train_features, train_labels, test_features, test_labels, k, lamchoices, numiters,avgnum):
    numlams = np.shape(lamchoices)[0]
    train_errs = [0]*numlams
    train_reconerrs = [0]*numlams
    train_classerrs = [0]*numlams
    train_accs = [0]*numlams
    test_errs = [0]*numlams
    test_reconerrs = [0]*numlams
    test_classerrs = [0]*numlams
    test_accs = [0]*numlams

    for i in range(numlams):
        for j in range(avgnum):
            module = TrainTestSetEvaluation(train_features, train_labels, test_features, test_labels, k, lamchoices[i], numiters)
            train_model_error, train_acc, numiters, [test_err, test_reconerr, test_classerr, test_acc], S_test = module.tt_eval_kl_ssnmfmult()
            train_model_errors = train_model_error[0]
            train_errs[i] = train_errs[i]+train_model_errors[numiters-1]
            train_model_reconerrs = train_model_error[1]
            train_reconerrs[i] = train_reconerrs[i]+train_model_reconerrs[numiters-1]
            train_model_classerrs = train_model_error[2]
            train_classerrs[i] = train_classerrs[i]+train_model_classerrs[numiters-1]
            train_accs[i] = train_accs[i]+train_acc
            test_errs[i] = test_errs[i]+test_err
            test_reconerrs[i] = test_reconerrs[i]+test_reconerr
            test_classerrs[i] = test_classerrs[i]+test_classerr
            test_accs[i] = test_accs[i]+test_acc

    train_errs = [element/avgnum for element in train_errs]
    train_reconerrs = [element/avgnum for element in train_reconerrs]
    train_classerrs = [element/avgnum for element in train_classerrs]
    train_accs = [element/avgnum for element in train_accs]
    test_errs = [element/avgnum for element in test_errs]
    test_reconerrs = [element/avgnum for element in test_reconerrs]
    test_classerrs = [element/avgnum for element in test_classerrs]
    test_accs = [element/avgnum for element in test_accs]

    plot_util(lamchoices,train_errs,test_errs,train_reconerrs,test_reconerrs,train_classerrs,test_classerrs,train_accs,test_accs,'l_KLSSNMF','l')
    
def numiters_plots(train_features, train_labels, test_features, test_labels, k, lam, numiterschoices,avgnum):
    numnums = np.shape(numiterschoices)[0]
    train_errs = [0]*numnums
    train_reconerrs = [0]*numnums
    train_classerrs = [0]*numnums
    train_accs = [0]*numnums
    test_errs = [0]*numnums
    test_reconerrs = [0]*numnums
    test_classerrs = [0]*numnums
    test_accs = [0]*numnums

    for i in range(numnums):
        for j in range(avgnum):
            module = TrainTestSetEvaluation(train_features, train_labels, test_features, test_labels, k, lam, numiterschoices[i])
            train_model_error, train_acc, numiters, [test_err, test_reconerr, test_classerr, test_acc], S_test = module.tt_eval_ssnmfmult()
            train_model_errors = train_model_error[0]
            train_errs[i] = train_errs[i]+train_model_errors[numiters-1]
            train_model_reconerrs = train_model_error[1]
            train_reconerrs[i] = train_reconerrs[i]+train_model_reconerrs[numiters-1]
            train_model_classerrs = train_model_error[2]
            train_classerrs[i] = train_classerrs[i]+train_model_classerrs[numiters-1]
            train_accs[i] = train_accs[i]+train_acc
            test_errs[i] = test_errs[i]+test_err
            test_reconerrs[i] = test_reconerrs[i]+test_reconerr
            test_classerrs[i] = test_classerrs[i]+test_classerr
            test_accs[i] = test_accs[i]+test_acc

    train_errs = [element/avgnum for element in train_errs]
    train_reconerrs = [element/avgnum for element in train_reconerrs]
    train_classerrs = [element/avgnum for element in train_classerrs]
    train_accs = [element/avgnum for element in train_accs]
    test_errs = [element/avgnum for element in test_errs]
    test_reconerrs = [element/avgnum for element in test_reconerrs]
    test_classerrs = [element/avgnum for element in test_classerrs]
    test_accs = [element/avgnum for element in test_accs]

    plot_util(numiterschoices,train_errs,test_errs,train_reconerrs,test_reconerrs,train_classerrs,test_classerrs,train_accs,test_accs,'n_SSNMF','n')

def kl_numiters_plots(train_features, train_labels, test_features, test_labels, k, lam, numiterschoices,avgnum):
    numnums = np.shape(numiterschoices)[0]
    train_errs = [0]*numnums
    train_reconerrs = [0]*numnums
    train_classerrs = [0]*numnums
    train_accs = [0]*numnums
    test_errs = [0]*numnums
    test_reconerrs = [0]*numnums
    test_classerrs = [0]*numnums
    test_accs = [0]*numnums

    for i in range(numnums):
        for j in range(avgnum):
            module = TrainTestSetEvaluation(train_features, train_labels, test_features, test_labels, k, lam, numiterschoices[i])
            train_model_error, train_acc, numiters, [test_err, test_reconerr, test_classerr, test_acc], S_test = module.tt_eval_kl_ssnmfmult()
            train_model_errors = train_model_error[0]
            train_errs[i] = train_errs[i]+train_model_errors[numiters-1]
            train_model_reconerrs = train_model_error[1]
            train_reconerrs[i] = train_reconerrs[i]+train_model_reconerrs[numiters-1]
            train_model_classerrs = train_model_error[2]
            train_classerrs[i] = train_classerrs[i]+train_model_classerrs[numiters-1]
            train_accs[i] = train_accs[i]+train_acc
            test_errs[i] = test_errs[i]+test_err
            test_reconerrs[i] = test_reconerrs[i]+test_reconerr
            test_classerrs[i] = test_classerrs[i]+test_classerr
            test_accs[i] = test_accs[i]+test_acc

    train_errs = [element/avgnum for element in train_errs]
    train_reconerrs = [element/avgnum for element in train_reconerrs]
    train_classerrs = [element/avgnum for element in train_classerrs]
    train_accs = [element/avgnum for element in train_accs]
    test_errs = [element/avgnum for element in test_errs]
    test_reconerrs = [element/avgnum for element in test_reconerrs]
    test_classerrs = [element/avgnum for element in test_classerrs]
    test_accs = [element/avgnum for element in test_accs]

    plot_util(numiterschoices,train_errs,test_errs,train_reconerrs,test_reconerrs,train_classerrs,test_classerrs,train_accs,test_accs,'n_KLSSNMF','n')



# Do not edit this string manually, always use bumpversion
# Details in CONTRIBUTING.md
__version__ = "1.0.3"

__author__ = "Jamie Haddock"
__email__ = "jhaddock@math.ucla.edu"

__license__ = "MIT"
__copyright__ = "Copyright (c) 2020, Jamie Haddock"

from .ssnmf import SSNMF



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
from numpy import linalg as la
import torch


class SSNMF_N:
    """
    Class for (S)NMF model.

    The NMF model consists of the data matrix to be factorized, X, the factor matrices, A and
    S.  Each model also consists of a label matrix, Y, classification factor matrix, B, and
    classification weight parameter, lam (although these three variables will be empty if Y is not
    input).  These parameters define the objective function defining the model:
    (1) ||X - AS||_F^2 or
    (2) D(X||AS) or
    (3) ||X - AS||_F^2 + lam * ||Y - BS||_F^2 or
    (4) ||X - AS||_F^2 + lam * D(Y||BS) or
    (5) D(X||AS) + lam * ||Y - BS||_F^2 or
    (6) D(X||AS) + lam * D(Y||BS).
    ...
    Parameters
    ----------
    X        : array
               Data matrix of size m x n.
    k        : int_
               Number of topics.
    modelNum : int_, optional
               Number indicating which of above models user intends to train (the default is 1).
    A        : array, optional
               Initialization for left factor matrix of X of size m x k (the default is a matrix with
               uniform random entries).
    S        : array, optional
               Initialization for right factor matrix of X of size k x n (the default is a matrix with
               uniform random entries).
    Y        : array, optional
               Label matrix of size p x n (default is None).
    B        : array, optional
               Initialization for left factor matrix of Y of size p x k (the default is a matrix with
               uniform random entries if Y is not None, None otherwise).
    lam      : float_, optional
               Weight parameter for classification term in objective (the default is 1 if Y is not
               None, None otherwise).
    W        : array, optional
               Missing data indicator matrix of same size as X (the defaults is matrix of all ones).
    L        : array, optional
               Missing label indicator matrix of same size as Y (the default is matrix of all ones if
               Y is not None, None otherwise).
    tol      : tolerance for termanating the model

    Methods
    -------
    mult(numiters = 10, saveerrs = True)
        Train the selected model via numiters multiplicative updates.
    accuracy()
        Compute the classification accuracy of supervised model (using Y, B, and S).
    kldiv()
        Compute the KL-divergence, D(Y||BS), of supervised model (using Y, B, and S).
    """

    def __init__(self, X, k, **kwargs):
        self.X = X
        rows = np.shape(X)[0]
        cols = np.shape(X)[1]
        self.modelNum = kwargs.get('modelNum', 1)  # initialize model indicator
        self.W = kwargs.get('W', np.ones((rows, cols)))  # initialize missing data indicator matrix
        self.A = kwargs.get('A', np.random.rand(rows, k))  # initialize factor A
        self.S = kwargs.get('S', np.random.rand(k, cols))  # initialize factor S
        self.tol = kwargs.get('tol', 1e-4) # initialize factor tol

        # check dimensions of X, A, and S match
        if rows != np.shape(self.A)[0]:
            raise Exception('The row dimensions of X and A are not equal.')
        if cols != np.shape(self.S)[1]:
            raise Exception('The column dimensions of X and S are not equal.')
        if np.shape(self.A)[1] != k:
            raise Exception('The column dimension of A is not k.')
        if np.shape(self.S)[0] != k:
            raise Exception('The row dimension of S is not k.')

        # supervision initializations (optional)
        self.Y = kwargs.get('Y', None)
        if self.Y is not None:
            # check dimensions of X and Y match
            if np.shape(self.Y)[1] != np.shape(self.X)[1]:
                raise Exception('The column dimensions of X and Y are not equal.')

            classes = np.shape(self.Y)[0]
            self.B = kwargs.get('B', np.random.rand(classes, k))
            self.lam = kwargs.get('lam', 1)
            self.L = kwargs.get('L', np.ones((classes, cols)))  # initialize missing label indicator matrix

            # check dimensions of Y, S, and B match
            if np.shape(self.B)[0] != classes:
                raise Exception('The row dimensions of Y and B are not equal.')
            if np.shape(self.B)[1] != k:
                raise Exception('The column dimension of B is not k.')
        else:
            self.B = None
            self.lam = None
            self.L = None

    def mult(self, **kwargs):
        '''
        Multiplicative updates for training (SS)NMF model.
        Parameters
        ----------
        numiters : int_, optional
            Number of iterations of updates to run (default is 10).
        saveerrs : bool, optional
            Boolean indicating whether to save model errors during iterations.
        eps : float_, optional
            Epsilon value to prevent division by zero (default is 1e-10).
        Returns
        -------
        errs : array, optional
            If saveerrs, returns array of ||X - AS||_F for each iteration (length numiters).
        '''
        numiters = kwargs.get('numiters', 10)
        saveerrs = kwargs.get('saveerrs', False)
        eps = kwargs.get('eps', 1e-10)
        initialErr = 0
        previousErr = 0
        currentErr = 0

        if saveerrs:
        # based on model number, initialize correct type of error array(s)
            if self.modelNum == 1 or self.modelNum == 2:
                errs = []  # initialize error array
            else :
                errs = []  # initialize error array
                reconerrs = []
                classerrs = []
                classaccs = []


        for i in range(numiters):
            # multiplicative updates for A, S, and possibly B
            # based on model number, use proper update functions for A,S,(B)
            if self.modelNum == 1:
                self.A = self.dictupdateFro(self.X, self.A, self.S, self.W, eps)
                self.S = np.transpose(self.dictupdateFro(np.transpose(self.X), np.transpose(self.S), np.transpose(self.A), np.transpose(self.W), eps))

                previousErr = currentErr
                currentErr = self.fronorm(self.X, self.A, self.S, self.W)
                if i == 0:
                    initialErr = currentErr

            if self.modelNum == 2:
                self.A = self.dictupdateIdiv(self.X, self.A, self.S, self.W, eps)
                self.S = np.transpose(self.dictupdateIdiv(np.transpose(self.X), np.transpose(self.S), np.transpose(self.A), np.transpose(self.W), eps))

                previousErr = currentErr
                currentErr = self.Idiv(self.X, self.A, self.S, self.W)
                if i == 0:
                    initialErr = currentErr

            if self.modelNum == 3:
                self.A = self.dictupdateFro(self.X, self.A, self.S, self.W, eps)
                self.B = self.dictupdateFro(self.Y, self.B, self.S, self.L, eps)
                self.S = self.repupdateFF(eps)

                previousErr = currentErr
                currentErr = self.fronorm(self.X, self.A, self.S, self.W)**2 + self.lam * (self.fronorm(self.Y, self.B, self.S, self.L)**2)
                if i == 0:
                    initialErr = currentErr

            if self.modelNum == 4:
                self.A = self.dictupdateFro(self.X, self.A, self.S, self.W, eps)
                self.B = self.dictupdateIdiv(self.Y, self.B, self.S, self.L, eps)
                self.S = self.repupdateIF(eps, self.modelNum)

                previousErr = currentErr
                currentErr = self.fronorm(self.X, self.A, self.S, self.W)**2 + self.lam * self.Idiv(self.Y, self.B, self.S, self.L)
                if i == 0:
                    initialErr = currentErr

            if self.modelNum == 5:
                self.A = self.dictupdateIdiv(self.X, self.A, self.S, self.W, eps)
                self.B = self.dictupdateFro(self.Y, self.B, self.S, self.L, eps)
                self.S = self.repupdateIF(eps, self.modelNum)

                previousErr = currentErr
                currentErr = self.Idiv(self.X, self.A, self.S, self.W) + self.lam * (self.fronorm(self.Y, self.B, self.S, self.L)**2)
                if i == 0:
                    initialErr = currentErr

            if self.modelNum == 6:
                self.A = self.dictupdateIdiv(self.X, self.A, self.S, self.W, eps)
                self.B = self.dictupdateIdiv(self.Y, self.B, self.S, self.L, eps)
                self.S = self.repupdateII(eps)

                previousErr = currentErr
                currentErr = self.Idiv(self.X, self.A, self.S, self.W) + self.lam * self.Idiv(self.Y, self.B, self.S, self.L)
                if i == 0:
                    initialErr = currentErr

            if i>0 and (previousErr - currentErr) / initialErr < self.tol:
                break

            if saveerrs:
            # based on model number, initialize correct type of error array(s)
                if self.modelNum == 1:
                    errs.append(self.fronorm(self.X, self.A, self.S, self.W))
                if self.modelNum == 2:
                    errs.append(self.Idiv(self.X, self.A, self.S, self.W))
                if self.modelNum == 3:
                    reconerrs.append(self.fronorm(self.X, self.A, self.S, self.W))
                    classerrs.append(self.fronorm(self.Y, self.B, self.S, self.L))
                    errs.append(reconerrs[i] ** 2 + self.lam * classerrs[i] ** 2)
                    classaccs.append(self.accuracy())
                if self.modelNum == 4:
                    reconerrs.append(self.fronorm(self.X, self.A, self.S, self.W))
                    classerrs.append(self.Idiv(self.Y, self.B, self.S, self.L))
                    errs.append(reconerrs[i] ** 2 + self.lam * classerrs[i])
                    classaccs.append(self.accuracy())
                if self.modelNum == 5:
                    reconerrs.append(self.Idiv(self.X, self.A, self.S, self.W))
                    classerrs.append(self.fronorm(self.Y, self.B, self.S, self.L))
                    errs.append(reconerrs[i] + self.lam * (classerrs[i] ** 2))  # save errors
                    classaccs.append(self.accuracy())
                if self.modelNum == 6:
                    reconerrs.append(self.Idiv(self.X, self.A, self.S, self.W))
                    classerrs.append(self.Idiv(self.Y, self.B, self.S, self.L))
                    errs.append(reconerrs[i] + self.lam * classerrs[i])  # save errors
                    classaccs.append(self.accuracy())

        if saveerrs:
            if self.modelNum == 1 or self.modelNum == 2:
                errs =np.array(errs)
                return [errs]
            else:
                errs = np.array(errs)
                reconerrs = np.array(reconerrs)
                classerrs = np.array(classerrs)
                classaccs = np.array(classaccs)
                return [errs, reconerrs, classerrs, classaccs]

    # based on model number, return correct type of error array(s)

    def dictupdateFro(self, Z, D, R, M, eps):
        '''
        multiplicitive update for D and R in ||Z - DR||_F^2
        Parameters
        ----------
        Z   : array
              Data matrix.
        D   : array
              Left factor matrix of Z.
        R   : array
              Right factor matrix of Z.
        M   : array
              Missing data indicator matrix of same size as Z (the defaults is matrix of all ones).
        eps : float_, optional
            Epsilon value to prevent division by zero (default is 1e-10).
        Returns
        -------
        updated D or the transpose of updated R
        '''

        return  np.multiply(
                np.divide(D, eps + np.multiply(M, D@R) @ np.transpose(R)), \
                np.multiply(M, Z) @ np.transpose(R))

    def dictupdateIdiv(self, Z, D, R, M, eps):
        '''
        multiplicitive update for D and R in D(Z||DR)
        Parameters
        ----------
        Z   : array
              Data matrix.
        D   : array
              Left factor matrix of Z.
        R   : array
              Right factor matrix of Z.
        M   : array
              Missing data indicator matrix of same size as Z (the defaults is matrix of all ones).
        eps : float_, optional
            Epsilon value to prevent division by zero (default is 1e-10).
        Returns
        -------
        updated D or the transpose of updated R
        '''
        return np.multiply(np.divide(D, eps + M @ np.transpose(R)), \
                           np.multiply(np.divide(np.multiply(M, Z), eps + np.multiply(M, D @ R)), M) @ np.transpose(R))

    def repupdateFF(self, eps):
    # update to use for S in model (3)
        return np.multiply(
            np.divide(self.S, eps + np.transpose(self.A) @ np.multiply(self.W, self.A @ self.S) + \
                      self.lam * np.transpose(self.B) @ np.multiply(self.L, self.B @ self.S)),
            np.transpose(self.A) @ np.multiply(self.W, self.X) + self.lam * np.transpose(self.B) @ np.multiply(self.L, self.Y)
        )

    def repupdateIF(self, eps, modelNum):
    # update to use for S in models (4) and (5)
        if modelNum == 4:
            return np.multiply(
                np.divide(self.S, eps + (2 * np.transpose(self.A) @ np.multiply(self.W, self.A @ self.S) +
                                         self.lam * np.transpose(self.B) @ self.L)),
                2 * np.transpose(self.A) @ np.multiply(self.W, self.X) + self.lam * np.transpose(self.B) @ \
                np.multiply(np.divide(np.multiply(self.L, self.Y), eps + np.multiply(self.L, self.B @ self.S)), self.L)
            )

        if modelNum == 5:
            return np.multiply(
                np.divide(self.S, eps + np.transpose(self.A) @ self.W + \
                                  2 * self.lam * np.transpose(self.B) @ np.multiply(self.L, self.B @ self.S)),
                np.transpose(self.A) @  np.multiply(np.divide(np.multiply(self.W, self.X), eps + np.multiply(self.W, self.A @ self.S)), self.W) + \
                2 * self.lam * np.transpose(self.B) @ np.multiply(self.L, self.Y)
            )

    def repupdateII(self, eps):
    # update to use for S in model (6)
        return np.multiply(
            np.divide(self.S, eps + np.transpose(self.A) @ self.W + self.lam * np.transpose(self.B) @ self.L),
            np.transpose(self.A) @ np.multiply(np.divide(np.multiply(self.W, self.X), eps + np.multiply(self.W, self.A @ self.S)), self.W) + \
            self.lam * np.transpose(self.B) @ np.multiply(np.divide(np.multiply(self.L, self.Y), eps + np.multiply(self.L, self.B @ self.S)), self.L)
        )

    def accuracy(self, **kwargs):
        '''
        Compute accuracy of supervised model.
        Parameters
        ----------
        Y : array, optional
            Label matrix (default is self.Y).
        B : array, optional
            Left factor matrix of Y (default is self.B).
        S : array, optional
            Right factor matrix of Y (default is self.S).
        L :
        Returns
        -------
        acc : float_
            Fraction of correctly classified data points (computed with Y, B, S).
        '''

        Y = kwargs.get('Y', self.Y)
        B = kwargs.get('B', self.B)
        S = kwargs.get('S', self.S)

        if Y is None:
            raise Exception('Label matrix Y not provided: model is not supervised.')

        numdata = np.shape(Y)[1]

        # count number of data points which are correctly classified
        numacc = 0
        Yhat = B @ S
        for i in range(numdata):
            true_max = np.argmax(Y[:, i])
            approx_max = np.argmax(Yhat[:, i])

            if true_max == approx_max:
                numacc = numacc + 1

        # return fraction of correctly classified data points
        acc = numacc / numdata
        return acc

    def Idiv(self, Z, D, S, M, **kwargs):
        '''
        Compute I-divergence between Z and DS.
        Parameters
        ----------
        Z   : array
              Data matrix.
        D   : array
              Left factor matrix of Z.
        S   : array
              Right factor matrix of Z.
        M   : array
              Missing data indicator matrix of same size as Z (the defaults is matrix of all ones).
        eps : float_, optional
            Epsilon value to prevent division by zero (default is 1e-10).
        Returns
        -------
        Idiv : float_
            I-divergence between Z and DS.
        '''
        eps = kwargs.get('eps', 1e-10)

        if Z is None:
            raise Exception('Matrix Z not provided.')
        if M is None:
            M = np.ones(Z.shape, dtype = float)

        # compute divergence
        Zhat = np.multiply(M, D @ S)
        div = np.multiply(np.multiply(M, Z), np.log(np.divide(np.multiply(M, Z) + eps, Zhat + eps))) \
              - np.multiply(M, Z) + Zhat
        Idiv = np.sum(np.sum(div))
        return Idiv

    def fronorm(self, Z, D, S, M, **kwargs):
        '''
        Compute Frobenius norm between Z and DS.
        Parameters
        ----------
        Z   : array
              Data matrix.
        D   : array
              Left factor matrix of Z.
        S   : array
              Right factor matrix of Z.
        M   : array
              Missing data indicator matrix of same size as Z (the defaults is matrix of all ones).
        Returns
        -------
        fronorm : float_
            Frobenius norm between Z and DS.
        '''

        if Z is None:
            raise Exception('Matrix Z not provided.')
        if M is None:
            M = np.ones(Z.shape, dtype = float)

        # compute norm
        Zhat = np.multiply(M, D @ S)
        fronorm = np.linalg.norm(np.multiply(M, Z) - Zhat, 'fro')
        return fronorm



class SSNMF_T:
    """
    Class for (S)NMF model.

    The NMF model consists of the data matrix to be factorized, X, the factor matrices, A and
    S.  Each model also consists of a label matrix, Y, classification factor matrix, B, and
    classification weight parameter, lam (although these three variables will be empty if Y is not
    input).  These parameters define the objective function defining the model:
    (1) ||X - AS||_F^2 or
    (2) D(X||AS) or
    (3) ||X - AS||_F^2 + lam * ||Y - BS||_F^2 or
    (4) ||X - AS||_F^2 + lam * D(Y||BS) or
    (5) D(X||AS) + lam * ||Y - BS||_F^2 or
    (6) D(X||AS) + lam * D(Y||BS).
    ...
    Parameters
    ----------
    X        : torch.Tensor
               Data matrix of size m x n.
    k        : int_
               Number of topics.
    modelNum : int_, optional
               Number indicating which of above models user intends to train (the default is 1).
    A        : torch.Tensor, optional
               Initialization for left factor matrix of X of size m x k (the default is a matrix with
               uniform random entries).
    S        : torch.Tensor, optional
               Initialization for right factor matrix of X of size k x n (the default is a matrix with
               uniform random entries).
    Y        : torch.Tensor, optional
               Label matrix of size p x n (default is None).
    B        : torch.Tensor, optional
               Initialization for left factor matrix of Y of size p x k (the default is a matrix with
               uniform random entries if Y is not None, None otherwise).
    lam      : float_, optional
               Weight parameter for classification term in objective (the default is 1 if Y is not
               None, None otherwise).
    W        : torch.Tensor, optional
               Missing data indicator matrix of same size as X (the defaults is matrix of all ones).
    L        : torch.Tensor, optional
               Missing label indicator matrix of same size as Y (the default is matrix of all ones if
               Y is not None, None otherwise).
    tol      : tolerance for termanating the model.

    Methods
    -------
    mult(numiters = 10, saveerrs = True)
        Train the selected model via numiters multiplicative updates.
    accuracy()
        Compute the classification accuracy of supervised model (using Y, B, and S).
    kldiv()
        Compute the KL-divergence, D(Y||BS), of supervised model (using Y, B, and S).
    """

    def __init__(self, X, k, **kwargs):
        cuda = torch.device('cuda') if torch.cuda.is_available() else torch.device("cpu")
        self.X = X
        rows = X.size(0)
        cols = X.size(1)
        self.modelNum = kwargs.get('modelNum', 1)  # initialize model indicator
        self.W = kwargs.get('W', torch.ones([rows, cols], dtype=torch.float, device=cuda))  # initialize missing data indicator matrix
        self.A = kwargs.get('A', torch.rand(rows, k, dtype=torch.float, device=cuda))  # initialize factor A
        self.S = kwargs.get('S', torch.rand(k, cols, dtype=torch.float, device=cuda))  # initialize factor S
        self.tol = kwargs.get('tol', 1e-4)  # initialize factor tol

        # check dimensions of X, A, and S match
        if rows != self.A.size(0):
            raise Exception('The row dimensions of X and A are not equal.')
        if cols != self.S.size(1):
            raise Exception('The column dimensions of X and S are not equal.')
        if self.A.size(1) != k:
            raise Exception('The column dimension of A is not k.')
        if self.S.size(0) != k:
            raise Exception('The row dimension of S is not k.')

        # supervision initializations (optional)
        self.Y = kwargs.get('Y', None)
        if self.Y is not None:
            # check dimensions of X and Y match
            if self.Y.size(1) != self.X.size(1):
                raise Exception('The column dimensions of X and Y are not equal.')

            classes = self.Y.size(0)
            self.B = kwargs.get('B', torch.rand(classes, k, dtype=torch.float, device=cuda))
            self.lam = kwargs.get('lam', 1)
            self.L = kwargs.get('L', torch.ones([classes, cols], dtype=torch.float, device=cuda))  # initialize missing label indicator matrix

            # check dimensions of Y, S, and B match
            if self.B.size(0) != classes:
                raise Exception('The row dimensions of Y and B are not equal.')
            if self.B.size(1) != k:
                raise Exception('The column dimension of B is not k.')
        else:
            self.B = None
            self.lam = None
            self.L = None

    def mult(self, **kwargs):
        '''
        Multiplicative updates for training (SS)NMF model.
        Parameters
        ----------
        numiters : int_, optional
            Number of iterations of updates to run (default is 10).
        saveerrs : bool, optional
            Boolean indicating whether to save model errors during iterations.
        eps : float_, optional
            Epsilon value to prevent division by zero (default is 1e-10).
        Returns
        -------
        errs : array, optional
            If saveerrs, returns array of ||X - AS||_F for each iteration (length numiters).
        '''
        numiters = kwargs.get('numiters', 10)
        saveerrs = kwargs.get('saveerrs', False)
        eps = kwargs.get('eps', 1e-10)
        initialErr = 0
        previousErr = 0
        currentErr = 0

        if saveerrs:
        # based on model number, initialize correct type of error array(s)
            if self.modelNum == 1 or self.modelNum == 2:
                errs = []  # initialize error array
            else :
                errs = []  # initialize error array
                reconerrs = []
                classerrs = []
                classaccs = []


        for i in range(numiters):
            # multiplicative updates for A, S, and possibly B
            # based on model number, use proper update functions for A,S,(B)
            if self.modelNum == 1:
                self.A = self.dictupdateFro(self.X, self.A, self.S, self.W, eps)
                self.S = torch.t(self.dictupdateFro(torch.t(self.X), torch.t(self.S), torch.t(self.A), torch.t(self.W), eps))

                previousErr = currentErr
                currentErr = self.fronorm(self.X, self.A, self.S, self.W)
                if i == 0:
                    initialErr = currentErr

            if self.modelNum == 2:
                self.A = self.dictupdateIdiv(self.X, self.A, self.S, self.W, eps)
                self.S = torch.t(self.dictupdateIdiv(torch.t(self.X), torch.t(self.S), torch.t(self.A), torch.t(self.W), eps))

                previousErr = currentErr
                currentErr = self.Idiv(self.X, self.A, self.S, self.W)
                if i == 0:
                    initialErr = currentErr

            if self.modelNum == 3:
                self.A = self.dictupdateFro(self.X, self.A, self.S, self.W, eps)
                self.B = self.dictupdateFro(self.Y, self.B, self.S, self.L, eps)
                self.S = self.repupdateFF(eps)

                previousErr = currentErr
                currentErr = self.fronorm(self.X, self.A, self.S, self.W)**2 + self.lam * (self.fronorm(self.Y, self.B, self.S, self.L)**2)
                if i == 0:
                    initialErr = currentErr

            if self.modelNum == 4:
                self.A = self.dictupdateFro(self.X, self.A, self.S, self.W, eps)
                self.B = self.dictupdateIdiv(self.Y, self.B, self.S, self.L, eps)
                self.S = self.repupdateIF(eps, self.modelNum)

                previousErr = currentErr
                currentErr = self.fronorm(self.X, self.A, self.S, self.W)**2 + self.lam * self.Idiv(self.Y, self.B, self.S, self.L)
                if i == 0:
                    initialErr = currentErr

            if self.modelNum == 5:
                self.A = self.dictupdateIdiv(self.X, self.A, self.S, self.W, eps)
                self.B = self.dictupdateFro(self.Y, self.B, self.S, self.L, eps)
                self.S = self.repupdateIF(eps, self.modelNum)

                previousErr = currentErr
                currentErr = self.Idiv(self.X, self.A, self.S, self.W) + self.lam * (self.fronorm(self.Y, self.B, self.S, self.L)**2)
                if i == 0:
                    initialErr = currentErr

            if self.modelNum == 6:
                self.A = self.dictupdateIdiv(self.X, self.A, self.S, self.W, eps)
                self.B = self.dictupdateIdiv(self.Y, self.B, self.S, self.L, eps)
                self.S = self.repupdateII(eps)

                previousErr = currentErr
                currentErr = self.Idiv(self.X, self.A, self.S, self.W) + self.lam * self.Idiv(self.Y, self.B, self.S, self.L)
                if i == 0:
                    initialErr = currentErr

            if i > 0 and (previousErr - currentErr) / initialErr < self.tol:
                break

            if saveerrs:
            # based on model number, initialize correct type of error array(s)
                if self.modelNum == 1:
                    errs.append(self.fronorm(self.X, self.A, self.S, self.W))
                if self.modelNum == 2:
                    errs.append(self.Idiv(self.X, self.A, self.S, self.W))
                if self.modelNum == 3:
                    reconerrs.append(self.fronorm(self.X, self.A, self.S, self.W))
                    classerrs.append(self.fronorm(self.Y, self.B, self.S, self.L))
                    errs.append(reconerrs[i] ** 2 + self.lam * classerrs[i] ** 2)
                    classaccs.append(self.accuracy())
                if self.modelNum == 4:
                    reconerrs.append(self.fronorm(self.X, self.A, self.S, self.W))
                    classerrs.append(self.Idiv(self.Y, self.B, self.S, self.L))
                    errs.append(reconerrs[i] ** 2 + self.lam * classerrs[i])
                    classaccs.append(self.accuracy())
                if self.modelNum == 5:
                    reconerrs.append(self.Idiv(self.X, self.A, self.S, self.W))
                    classerrs.append(self.fronorm(self.Y, self.B, self.S, self.L))
                    errs.append(reconerrs[i] + self.lam * (classerrs[i] ** 2))  # save errors
                    classaccs.append(self.accuracy())
                if self.modelNum == 6:
                    reconerrs.append(self.Idiv(self.X, self.A, self.S, self.W))
                    classerrs.append(self.Idiv(self.Y, self.B, self.S, self.L))
                    errs.append(reconerrs[i] + self.lam * classerrs[i])  # save errors
                    classaccs.append(self.accuracy())

        if saveerrs:
            if self.modelNum == 1 or self.modelNum == 2:
                if torch.cuda.is_available():
                    errs = torch.cuda.FloatTensor(errs)
                else:
                    errs = torch.FloatTensor(errs)
                return [errs]
            else:
                if torch.cuda.is_available():
                    errs = torch.cuda.FloatTensor(errs)
                    reconerrs = torch.cuda.FloatTensor(reconerrs)
                    classerrs = torch.cuda.FloatTensor(classerrs)
                    classaccs = torch.cuda.FloatTensor(classaccs)
                else:
                    errs = torch.FloatTensor(errs)
                    reconerrs = torch.FloatTensor(reconerrs)
                    classerrs = torch.FloatTensor(classerrs)
                    classaccs = torch.FloatTensor(classaccs)
                return [errs, reconerrs, classerrs, classaccs]

    # based on model number, return correct type of error array(s)

    def dictupdateFro(self, Z, D, R, M, eps):
        '''
        multiplicitive update for D and R in ||Z - DR||_F^2
        Parameters
        ----------
        Z   : torch.Tensor
              Data matrix.
        D   : torch.Tensor
              Left factor matrix of Z.
        R   : torch.Tensor
              Right factor matrix of Z.
        M   : torch.Tensor
              Missing data indicator matrix of same size as Z (the defaults is matrix of all ones).
        eps : float_, optional
            Epsilon value to prevent division by zero (default is 1e-10).
        Returns
        -------
        updated D or the transpose of updated R
        '''

        return torch.mul(
               torch.div(D, eps + torch.mul(M, D@R) @ torch.t(R)),
               torch.mul(M, Z) @ torch.t(R))

    def dictupdateIdiv(self, Z, D, R, M, eps):
        '''
        multiplicitive update for D and R in D(Z||DR)
        Parameters
        ----------
        Z   : torch.Tensor
              Data matrix.
        D   : torch.Tensor
              Left factor matrix of Z.
        R   : torch.Tensor
              Right factor matrix of Z.
        M   : torch.Tensor
              Missing data indicator matrix of same size as Z (the defaults is matrix of all ones).
        eps : float_, optional
            Epsilon value to prevent division by zero (default is 1e-10).
        Returns
        -------
        updated D or the transpose of updated R
        '''
        return torch.mul(torch.div(D, eps + M @ torch.t(R)),
                         torch.mul(torch.div(torch.mul(M, Z), eps + torch.mul(M, D @ R)), M) @ torch.t(R))

    def repupdateFF(self, eps):
    # update to use for S in model (3)
        return torch.mul(
            torch.div(self.S, eps + torch.t(self.A) @ torch.mul(self.W, self.A @ self.S) +
                      self.lam * torch.t(self.B) @ torch.mul(self.L, self.B @ self.S)),
            torch.t(self.A) @ torch.mul(self.W, self.X) + self.lam * torch.t(self.B)
            @ torch.mul(self.L, self.Y))

    def repupdateIF(self, eps, modelNum):
    # update to use for S in models (4) and (5)
        if modelNum == 4:
            return torch.mul(
                torch.div(self.S, eps + (2 * torch.t(self.A) @ torch.mul(self.W, self.A @ self.S) +
                                         self.lam * torch.t(self.B) @ self.L)
                          ),
                2 * torch.t(self.A) @ torch.mul(self.W, self.X) + self.lam * torch.t(self.B) @ \
                torch.mul(torch.div(torch.mul(self.L, self.Y), eps + torch.mul(self.L, self.B @ self.S)), self.L)
            )

        if modelNum == 5:
            return torch.mul(
                torch.div(self.S, eps + torch.t(self.A) @ self.W + 2 * self.lam * torch.t(self.B) @ torch.mul(self.L, self.B @ self.S)
                          ),
                torch.t(self.A) @ torch.mul(torch.div(torch.mul(self.W, self.X), eps + torch.mul(self.W, self.A @ self.S)), self.W) + \
                2 * self.lam * torch.t(self.B) @ torch.mul(self.L, self.Y)
            )

    def repupdateII(self, eps):
    # update to use for S in model (6)
        return torch.mul(
            torch.div(self.S, eps + torch.t(self.A) @ self.W + self.lam * torch.t(self.B) @ self.L),
            torch.t(self.A) @ torch.mul(torch.div(torch.mul(self.W, self.X), eps + torch.mul(self.W, self.A @ self.S)), self.W) + \
            self.lam * torch.t(self.B) @ torch.mul(torch.div(torch.mul(self.L, self.Y), eps + torch.mul(self.L, self.B @ self.S)), self.L)
        )

    def accuracy(self, **kwargs):
        '''
        Compute accuracy of supervised model.
        Parameters
        ----------
        Y : torch.tensor, optional
            Label matrix (default is self.Y).
        B : torch.tensor, optional
            Left factor matrix of Y (default is self.B).
        S : torch.tensor, optional
            Right factor matrix of Y (default is self.S).
        Returns
        -------
        acc : float_
            Fraction of correctly classified data points (computed with Y, B, S).
        '''

        Y = kwargs.get('Y', self.Y)
        B = kwargs.get('B', self.B)
        S = kwargs.get('S', self.S)

        if Y is None:
            raise Exception('Label matrix Y not provided: model is not supervised.')

        numdata = Y.size(1)

        # count number of data points which are correctly classified
        numacc = 0
        Yhat = B @ S
        for i in range(numdata):
            true_max = torch.argmax(Y[:, i])
            approx_max = torch.argmax(Yhat[:, i])

            if true_max == approx_max:
                numacc = numacc + 1

        # return fraction of correctly classified data points
        acc = numacc / numdata
        return acc

    def Idiv(self, Z, D, S, M, **kwargs):
        '''
        Compute I-divergence between Z and DS.
        Parameters
        ----------
        Z   : torch.tensor
              Data matrix.
        D   : torch.tensor
              Left factor matrix of Z.
        S   : torch.tensor
              Right factor matrix of Z.
        M   : torch.tensor
              Missing data indicator matrix of same size as Z (the defaults is matrix of all ones).
        eps : float_, optional
            Epsilon value to prevent division by zero (default is 1e-10).
        Returns
        -------
        Idiv : float_
            I-divergence between Z and DS.
        '''
        eps = kwargs.get('eps', 1e-10)

        if Z is None:
            raise Exception('Matrix Z not provided.')
        if M is None:
            M = torch.ones(Z.size(0),Z.size(1), dtype=torch.float, device = Z.device)

        # compute divergence
        Zhat = torch.mul(M, D @ S)
        div = torch.mul(torch.mul(M, Z), torch.log(torch.div(torch.mul(M, Z) + eps, Zhat + eps))) \
              - torch.mul(M, Z) + Zhat
        Idiv = torch.sum(torch.sum(div))
        return Idiv

    def fronorm(self, Z, D, S, M, **kwargs):
        '''
        Compute Frobenius norm between Z and DS.
        Parameters
        ----------
        Z   : torch.tensor
              Data matrix.
        D   : torch.tensor
              Left factor matrix of Z.
        S   : torch.tensor
              Right factor matrix of Z.
        M   : torch.tensor
              Missing data indicator matrix of same size as Z (the defaults is matrix of all ones).
        Returns
        -------
        fronorm : float_
            Frobenius norm between Z and DS.
        '''

        if Z is None:
            raise Exception('Matrix Z not provided.')
        if M is None:
            M = torch.ones(Z.size(0), Z.size(1), dtype=torch.float, device=Z.device)

        # compute norm
        Zhat = torch.mul(M, D @ S)
        fronorm = torch.norm(torch.mul(M, Z) - Zhat, p='fro')
        return fronorm



class SSNMF(SSNMF_N, SSNMF_T):
    """
        Class for (S)NMF model.
        This class is inherited from class SSNMF_N and class SSNMF_T. It can take both Numpy array and
        PyTorch tensor when initializing the model.

        The NMF model consists of the data matrix to be factorized, X, the factor matrices, A and
        S.  Each model also consists of a label matrix, Y, classification factor matrix, B, and
        classification weight parameter, lam (although these three variables will be empty if Y is not
        input).  These parameters define the objective function defining the model:
        (1) ||X - AS||_F^2 or
        (2) D(X||AS) or
        (3) ||X - AS||_F^2 + lam * ||Y - BS||_F^2 or
        (4) ||X - AS||_F^2 + lam * D(Y||BS) or
        (5) D(X||AS) + lam * ||Y - BS||_F^2 or
        (6) D(X||AS) + lam * D(Y||BS).
        ...
        Parameters
        ----------
        X        : torch.Tensor
                   Data matrix of size m x n.
        k        : int_
                   Number of topics.
        modelNum : int_, optional
                   Number indicating which of above models user intends to train (the default is 1).
        A        : torch.Tensor, optional
                   Initialization for left factor matrix of X of size m x k (the default is a matrix with
                   uniform random entries).
        S        : torch.Tensor, optional
                   Initialization for right factor matrix of X of size k x n (the default is a matrix with
                   uniform random entries).
        Y        : torch.Tensor, optional
                   Label matrix of size p x n (default is None).
        B        : torch.Tensor, optional
                   Initialization for left factor matrix of Y of size p x k (the default is a matrix with
                   uniform random entries if Y is not None, None otherwise).
        lam      : float_, optional
                   Weight parameter for classification term in objective (the default is 1 if Y is not
                   None, None otherwise).
        W        : torch.Tensor, optional
                   Missing data indicator matrix of same size as X (the defaults is matrix of all ones).
        L        : torch.Tensor, optional
                   Missing label indicator matrix of same size as Y (the default is matrix of all ones if
                   Y is not None, None otherwise).
        tol      : Tolerance for relative error stopping criterion
                  (i.e., method stops when difference between consecutive relative errors falls below tol) 
        str      : a flag to indicate whether this model is initialized by Numpy array or PyTorch tensor

        Methods
        -------
        mult(numiters = 10, saveerrs = True)
            Train the selected model via numiters multiplicative updates.
        accuracy()
            Compute the classification accuracy of supervised model (using Y, B, and S).
        kldiv()
            Compute the KL-divergence, D(Y||BS), of supervised model (using Y, B, and S).
        """

    def __init__(self, X, k, **kwargs):
        if type(X) == np.ndarray:
            SSNMF_N.__init__(self, X, k, **kwargs)
            self.str = "numpy"  # protected attribute
        if type(X) == torch.Tensor:
            SSNMF_T.__init__(self, X, k, **kwargs)
            self.str = "torch"  # protected attribute

    def mult(self, **kwargs):
        '''
        Multiplicative updates for training (SS)NMF model.
        Parameters
        ----------
        numiters : int_, optional
            Number of iterations of updates to run (default is 10).
        saveerrs : bool, optional
            Boolean indicating whether to save model errors during iterations.
        eps : float_, optional
            Epsilon value to prevent division by zero (default is 1e-10).
        Returns
        -------
        errs : Numpy array or PyTorch tensor, optional
            If saveerrs, returns array of ||X - AS||_F for each iteration (length numiters).
        '''

        if self.str == "numpy":
            return SSNMF_N.mult(self, **kwargs)
        if self.str == "torch":
            return SSNMF_T.mult(self, **kwargs)

    def accuracy(self, **kwargs):
        '''
        Compute accuracy of supervised model.
        Parameters
        ----------
        Y : Numpy array or PyTorch tensor, optional
            Label matrix (default is self.Y).
        B : Numpy array or PyTorch tensor, optional
            Left factor matrix of Y (default is self.B).
        S : Numpy array or PyTorch tensor, optional
            Right factor matrix of Y (default is self.S).
        L :
        Returns
        -------
        acc : float_
            Fraction of correctly classified data points (computed with Y, B, S).
        '''


        if self.str == "numpy":
            return SSNMF_N.accuracy(self, **kwargs)
        if self.str == "torch":
            return SSNMF_T.accuracy(self, **kwargs)

    def Idiv(self, Z, D, S, M, **kwargs):
        '''
        Compute I-divergence between Z and DS.
        Parameters
        ----------
        Z   : Numpy array or PyTorch tensor
              Data matrix.
        D   : Numpy array or PyTorch tensor
              Left factor matrix of Z.
        S   : Numpy array or PyTorch tensor
              Right factor matrix of Z.
        M   : Numpy array or PyTorch tensor
              Missing data indicator matrix of same size as Z (the defaults is matrix of all ones).
        eps : float_, optional
            Epsilon value to prevent division by zero (default is 1e-10).
        Returns
        -------
        Idiv : float_
            I-divergence between Z and DS.
        '''
        if self.str == "numpy":
            return SSNMF_N.Idiv(self, Z, D, S, M, **kwargs)
        if self.str == "torch":
            return SSNMF_T.Idiv(self, Z, D, S, M, **kwargs)

    def fronorm(self, Z, D, S, M, **kwargs):
        '''
        Compute Frobenius norm between Z and DS.
        Parameters
        ----------
        Z   : Numpy array or PyTorch tensor
              Data matrix.
        D   : Numpy array or PyTorch tensor
              Left factor matrix of Z.
        S   : Numpy array or PyTorch tensor
              Right factor matrix of Z.
        M   : Numpy array or PyTorch tensor
              Missing data indicator matrix of same size as Z (the defaults is matrix of all ones).
        Returns
        -------
        fronorm : float_
            Frobenius norm between Z and DS.
        '''
        if self.str == "numpy":
            return SSNMF_N.fronorm(self, Z, D, S, M, **kwargs)
        if self.str == "torch":
            return SSNMF_T.fronorm(self, Z, D, S, M, **kwargs)




